//
//  covalsymbol.c
//
//  coval version 0.8
//      2012.11.01 Takeshi Ogihara
//  Copyright (C) 2010, 2011, 2012 Takeshi Ogihara
//
#include <coval.h>
#include <covalsymbol.h>
#ifdef __USE_MALLOC__
#include <stdlib.h>
#ifdef __USE_STRDUP__
#include <string.h>
#endif
#endif

#ifndef COVAL_SYMBOL_TABLE_SIZE
#define COVAL_SYMBOL_TABLE_SIZE  32
#endif
#ifndef COVAL_SYMBOL_MAXLENGTH
#define COVAL_SYMBOL_MAXLENGTH   18
#endif

typedef struct _cov_symbol {
    coval_base *coval;
#ifdef __USE_MALLOC__
    const char *symbol;
#else
    char symbol[COVAL_SYMBOL_MAXLENGTH];
#endif
} cov_symbol_entry;

#ifndef __USE_MALLOC__
static cov_symbol_entry _deftable[COVAL_SYMBOL_TABLE_SIZE];
#endif
static unsigned int _tablesize;

static cov_symbol_entry *theTable = NULL;

#ifdef __USE_MALLOC__
#ifndef __USE_STRDUP__
static char *strdup(const char *str)
{
    char *p;
    unsigned int lng = 0;
    while (str[lng]) lng++;
    p = malloc(lng + 1);
    for (lng = 0; (p[lng] = str[lng]) != 0; lng++)
        ;
    return p;
}

static int strcmp(const char *p, const char *q) {
    int d;
    while ((d = *p - *q) == 0 && *p != 0)
        p++, q++;
    return d;
}
#endif /* __USE_STRDUP__ */
#else
static void str_cpy(char *buf, const char *str) {
    int i;
    for (i = 0; str[i] != 0 && i < COVAL_SYMBOL_MAXLENGTH - 1; i++)
        buf[i] = str[i];
    buf[i] = 0;
}

static int str_cmp(const char *p, const char *q) {
    int i = 0, d;
    while ((d = p[i] - q[i]) == 0 && p[i] != 0) {
        if (++i >= COVAL_SYMBOL_MAXLENGTH - 1)
            return 0;
    }
    return d;
}
#endif

int CovalSymbolTableInit(unsigned int capacity)
{
    if (theTable)
        return _tablesize;
#ifdef __USE_MALLOC__
    unsigned int cap = (capacity + 7) & ~07U;
    theTable = calloc(cap,  sizeof(cov_symbol_entry));
    _tablesize = cap;
#else
    theTable = _deftable;
    _tablesize = COVAL_SYMBOL_TABLE_SIZE;
#endif
    return _tablesize;
}

static unsigned int hash(const char *p) {
    unsigned int i, v = *p;
    unsigned int n;
    for (n = 1, i = 1; p[n]; ) {
        v += (p[n] << i);
        if (++i > 4) i = 0;
        n++;
#ifndef __USE_MALLOC__
        if (n >= COVAL_SYMBOL_MAXLENGTH - 1)
            break;
#endif
    }
    return v % _tablesize;
}

#define hashnext(x) (((x) + 17) % _tablesize)

coval_base *CovalSymbolTableGet(const char *sym)
{
    if (sym == NULL || *sym == 0)
        return 0;
    unsigned int x = hash(sym);
#ifdef __USE_MALLOC__
    while (theTable[x].symbol) {
        if (strcmp(theTable[x].symbol, sym) == 0)
            return theTable[x].coval;
        x = hashnext(x);
    }
#else
    char *p;
    while (*(p = theTable[x].symbol) != 0) {
        if (str_cmp(p, sym) == 0)
            return theTable[x].coval;
        x = hashnext(x);
    }
#endif
    return NULL;
}

void CovalSymbolTableSet(const char *sym, coval_base *cov)
{
    if (sym == NULL || *sym == 0)
        return; // Error
    unsigned int x = hash(sym);
    while (theTable[x].coval) {
        x = hashnext(x);
    }
    theTable[x].coval = cov;
#ifdef __USE_MALLOC__
    theTable[x].symbol = strdup(sym);
#else
    str_cpy(theTable[x].symbol, sym);
#endif
}

#ifdef __TEST_ALONE
#include <stdio.h>
int main(void) {
    long tag[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };
    CovalSymbolTableInit(48);
    CovalSymbolTableSet("TheFirst", (coval_base *)&tag[1]);
    CovalSymbolTableSet("TheSecond", (coval_base *)&tag[2]);
    CovalSymbolTableSet("3rd", (coval_base *)&tag[3]);
    CovalSymbolTableSet("The4thLongLongIdentifierForTheTest", (coval_base *)&tag[4]);
    long *p;
    p = (long *)CovalSymbolTableGet("None");
    printf("val=%ld\n", p ? *p : 0L);
    p = (long *)CovalSymbolTableGet("TheSecond");
    printf("val=%ld\n", p ? *p : 0L);
    p = (long *)CovalSymbolTableGet("The4thLongLongIdentifierForTheTeXX");
    printf("val=%ld\n", p ? *p : 0L);
    return 0;
}
#endif
